#!/usr/bin/perl -w

# parse information from XML SwissProt files
# Karsten Hokamp, Brinkman Lab, Simon Fraser University
# May 2004

use strict;
use Getopt::Long;

my $revision_number = '$Revision: 1.4 $'; # '  $Date: 2005/07/11 22:24:46 $
$revision_number =~ s/[\$Revision\: | \$]//g;

my $fh = *STDIN;
my $error_key = 'ERROR: key not found!';
my $file = '';
my $restrict_organisms = '';
my $restrict_organisms_regexp = '';
my %restrict_organisms = ();
my %restrict_organisms_regexp = ();
my $include_all = '';
my %skip_entries = ();
my %found_entries = ();

&GetOptions('org=s', \$restrict_organisms,
	    'orgre=s', \$restrict_organisms_regexp,
	    'all', \$include_all,
	    );
if ($restrict_organisms) {
    foreach (split /\,/, $restrict_organisms) {
	s/^\s*//;
	s/\s*$//;
	$restrict_organisms{$_}++;
    }
}
if ($restrict_organisms_regexp) {
    foreach (split /\,/, $restrict_organisms_regexp) {
#	s/^\s*//;
#	s/\s*$//;
	$restrict_organisms_regexp{$_}++;
    }
}

if (@ARGV
    and
    -r $ARGV[0]) {
    $file = shift @ARGV;
    print "# Processing file '$file'...\n";

    open (IN, $file) 
	or die;
    $fh = *IN;
}
    
my %global_info = ();
my $entry = '';
my $entry_counter = 0;
while (<$fh>) {
    
    # initialize entry with first line:
    $entry = $_;
    
    # add lines until the end of an entry
    until (/^<\/entry>/
	   or
	   eof($fh)) {
	$_ = <$fh>;
	$entry .= $_;
    }
    
    $entry_counter++;
    &parse($entry);
    
    # do something with %info
}

# print output:
print "#accessions\tname\torganism\tgene names\tprotein names\tfunctions\n";
foreach my $id (sort keys %global_info) {
    my @accession = ();
    my @function = ();
    my @gene_name = ();
    my @name = ();
    my @organism = ();
    my @protein_name = ();
    my @out = ();

    # collect data:
    if (defined $global_info{$id}{'accession'}) {
	@accession = @{$global_info{$id}{'accession'}};
    }
    push @out, join '; ', @accession;

    if (defined $global_info{$id}{'name'}) {
	@name = @{$global_info{$id}{'name'}};
    }
    push @out, join '; ', @name;

    if (defined $global_info{$id}{'organism'}) {
	@organism = @{$global_info{$id}{'organism'}};
    }
    push @out, join '; ', @organism;
    my $organism = join '; ', @organism;;
    $found_entries{$organism}++;

    if (defined $global_info{$id}{'gene_name'}) {
	@gene_name = @{$global_info{$id}{'gene_name'}};
    }
    push @out, join '; ', @gene_name;

    if (defined $global_info{$id}{'protein_name'}) {
	@protein_name = @{$global_info{$id}{'protein_name'}};
    }
    push @out, join '; ', @protein_name;

    if (defined $global_info{$id}{'function'}) {
	@function = @{$global_info{$id}{'function'}};
    }
    push @out, join '; ', @function;

    print (join "\t", @out);
    print "\n";
}


print STDERR "# Finished - $entry_counter entries processed!\n";
if (%found_entries) {
    print STDERR "\n# The following entries were found:\n";
    foreach (sort { $found_entries{$b} <=> $found_entries{$a} } keys %found_entries) {
	print STDERR "#\t $_: $found_entries{$_}\n";
    }
}
if (%skip_entries) {
    print STDERR "\n# The following entries were skipped:\n";
    foreach (sort { $skip_entries{$b} <=> $skip_entries{$a} } keys %skip_entries) {
	print STDERR "#\t $_: $skip_entries{$_}\n";
    }
}

##### END OF MAIN #####
#######################
##### SUBROUTINES #####

sub parse {
    my $entry = shift;
#    my %out = ();
    my $value = '';
    my @values = ();
    my $id = '';

    # get organism
    @values = &get_XML_entry(\$entry,'organism->name|type="scientific"');
    if (@values > 1) {
	die "# Multiple organisms found in $entry\n";
    }
    my $organism = $values[0];    
    if ($organism eq $error_key) {
	if (defined $restrict_organisms{'no organism'}) {
	    $organism = 'no organism';
	} else {
	    if ($include_all) {
		$organism = '';
	    } else {
		warn "No organism in $entry!\n";
		$skip_entries{'no organism'}++;
		return;
	    }
	}
    }
    
    my $found_org = 0;
    foreach (keys %restrict_organisms_regexp) {
	if ($organism =~ /$_/) {
	    $found_org = 1;
	    last;
	}
    }

    unless ($found_org) {
	if (%restrict_organisms) {
	    unless (defined $restrict_organisms{$organism}) {
		$skip_entries{$organism}++;
		return;
	    }
	}
    }

    # get accession
    @values = ();
    while ($entry) {
	($value) = &get_XML_entry(\$entry,'accession');
	if ($value =~ /$error_key/) {
	    last;
	} else {
	    push @values, $value;
	}
    }
    unless (@values) {
	warn "No accession in entry '$entry' - skipping\n";
	return ();
    }	
    if (@values > 1) {
	warn "# Multiple accession found: @values\n";
    }
    $value = join '; ', (sort @values);

    if (defined $global_info{$value}) {
	die "accession '$value' is defined multiple times!";
    }
    @{$global_info{$value}{'accession'}} = @values;

    # use accession as id for this entry
    $id = $value;

    push @{$global_info{$id}{'organism'}}, $organism;

    # get name(s)
    @values = &get_XML_entry(\$entry,'name');
    foreach (@values) {
	if ($_ eq $error_key) {
	    warn "No name for $id!\n";
	    next;
	}
	push @{$global_info{$id}{'name'}}, $_;
    }


    # get gene name(s)
    @values = &get_XML_entry(\$entry,'gene->name');
    foreach (@values) {
	if ($_ eq $error_key) {
#	    warn "No gene->name for $id!\n";
	    next;
	}
	push @{$global_info{$id}{'gene_name'}}, $_;
    }

    # get protein name(s)
    @values = &get_XML_entry(\$entry,'protein->name');
    foreach (@values) {
	if ($_ eq $error_key) {
#	    warn "No protein->name for $id!\n";
	    next;
	}
	push @{$global_info{$id}{'protein_name'}}, $_;
    }

    # get function(s)
    @values = &get_XML_entry(\$entry,'comment|type="function"->text');
    foreach (@values) {
	if ($_ eq $error_key) {
	    warn "No function for $id!\n";
	    next;
	}
	push @{$global_info{$id}{'function'}}, $_;
    }
}


sub get_XML_entry {
    # extract entry for key and remove processed field
    my $entry = shift;
    my $key = shift;

    my @keys = (split /->/, $key);

    my $first_key = shift @keys;
    my $qualifier = '';
    ($first_key,$qualifier) = split /\|/, $first_key;
    if (defined $qualifier) {
	$qualifier = ' '.$qualifier;
    } else {
	$qualifier = '';
    }

    my $info_section = '';
    unless ($$entry =~ /<$first_key$qualifier.*?>(.*?)<\/$first_key>/s) {
	return $error_key;
    } else {
	$$entry =~ s/<$first_key$qualifier.*?>(.*?)<\/$first_key>//s;
	$info_section = $1;
    }   
    
    my @out = ();
    unless (@keys) {
	@out = ($info_section);
    }

    foreach my $sub_key (@keys) {
	$qualifier = '';
	($sub_key,$qualifier) = split /\|/, $sub_key;
	if (defined $qualifier) {
	    $qualifier = ' '.$qualifier;
	} else {
	    $qualifier = '';
	}

	unless ($info_section =~ /<$sub_key$qualifier.*?>(.*?)<\/$sub_key>/s) {
	    return $error_key;
	} else {
	    while ($info_section =~ /<$sub_key$qualifier.*?>(.*?)<\/$sub_key>/s) {
		push @out, $1;
		$info_section =~ s/<$sub_key$qualifier.*?>(.*?)<\/$sub_key>//s;
	    }   
	}
    }

    return @out;
}
